
local error, type, pairs, ipairs, tostring = error, type, pairs, ipairs, tostring
local insert, sort, concat = table.insert, table.sort, table.concat
local exit = os.exit
local max, min = math.max, math.min
local char = string.char
local print = print

module ("nsg")

local optionsDescr
local function exitWithUsage(err)
  local s = usage(optionsDescr)
  print(err..s)
  exit(1)
end

local function err_unknown_opt (opt)
  exitWithUsage("Unknown option `-" ..
    (#opt > 1 and "-" or "") .. opt .. "'\n")
end

local function canonize (options, opt, noErr)
  if not options [opt] then
    if noErr then return nil end
    err_unknown_opt (opt)
  end
  local realOpt = opt
  while type (options [opt]) == "string" do
    opt = options [opt]
    if not options [opt] then
      if noErr then return nil end
      err_unknown_opt (realOpt)
    end
  end
  return opt
end

function usage(optsDescr)
  local usage = {}
  local hlp = {}
  for k, v in pairs(optsDescr) do
    if type(k) == 'string' then
      local opt = canonize(optsDescr, k, true)
      if opt then
        if not usage[opt] then usage[opt] = {aliases = {}} end
        insert(usage[opt].aliases, k)
        if type(v) == 'table' then
          for i = 1, #v do
            usage[opt][i] = v[i]
          end
        end
      end
    end
  end
  local maxoptsize = 0
  for k, v in pairs(usage) do
    for i, opt in ipairs(v.aliases) do
      v.aliases[i] = '\253'..(#opt > 1 and '\253' or '')..opt
    end
    sort(v.aliases)
    if #v.aliases[1] > 2 then
      v.aliases[1] = '\255\255\255\255'..v.aliases[1]
    end
    if v[1] ~= 0 then
      local anum = #v.aliases
      local opt = v.aliases[anum]:match"^\255*\253*(.*)$"

      v.aliases[anum] = v.aliases[anum]
      ..((#opt > 1 and '=' or ' ')..(v[3] or opt:upper()))
    end
    local s = concat(v.aliases,', ')
    maxoptsize = max(#s, maxoptsize)
    hlp[#hlp+1] = s..'\254'..char(#s)..(v[2] or '<no help>')
  end
  sort(hlp)
  return "\nUsage: "..concat(optsDescr, '\n')..'\n\n'
  ..concat(hlp, '\n'):gsub('\253', '-'):gsub('\255', ' ')
  :gsub('\254(.)', function(s)
      return (' '):rep(min(maxoptsize, 31) + 1 - s:byte())
  end)..'\n'
end

function getopt_ordered (arg, long_opts)
  optionsDescr = long_opts
  local i      = 1
  local count  = 1
  local opts   = {}
  local optarg = {}
  
  local options = {}
  for k,v in pairs (long_opts) do
    if type(v) == "table" then
      options[k] = v[1]
    elseif type(v) == "string" then
      options[k] = v
    else
      error("Bad options description: "..tostring(k))
    end
  end
  
  while i <= #arg do
    local a = arg [i]
    if a == "--" then
      i = i + 1
      break
    elseif a == "-" then
      break
    elseif a:sub (1, 2) == "--" then
      local pos = a:find ("=", 1, true)
      if pos then
        local opt = a:sub (3, pos-1)
        opt = canonize (options, opt)
        if options [opt] == 0 then
          exitWithUsage("Bad usage of option `" .. a .. "'\n")
        end
        optarg [count] = a:sub (pos+1)
        opts [count] = opt
      else
        local opt = a:sub (3)
        opt = canonize (options, opt)
        if options [opt] == 0 then
          opts [count] = opt
        else
          if i == #arg then
            exitWithUsage("Missed value for option `" .. a .. "'\n")
          end
          optarg [count] = arg [i+1]
          opts [count] = opt
          i = i + 1
        end
      end
      count = count + 1
    elseif a:sub (1, 1) == "-" then
      for j = 2, #a do
        local opt = canonize (options, a:sub (j, j))
        if options [opt] == 0 then
          opts [count] = opt
        elseif #a == j then
          if i == #arg then
            exitWithUsage("Missed value for option `" ..'-'.. a:sub(j,j) .. "'\n")
          end
          optarg [count] = arg [i+1]
          opts [count] = opt
          i = i + 1
          count = count + 1
          break
        else--if j == 2 then
          optarg [count] = a:sub (j + 1)
          opts [count] = opt
          count = count + 1
          break
          --else
          --exitWithUsage("Bad usage of option `-" .. opt .. "'\n")
        end
        count = count + 1
      end
    else
      optarg [count] = arg [i]
      opts [count] = false
      count = count + 1
    end
    i = i + 1
  end
  for k = i, #arg do
    optarg [count] = arg [k]
    opts [count] = false
    count = count + 1
  end
  return opts, optarg
end

function getopt(arg, long_opts, multi)
  local ret = {}
  local other = {}
  local opts, optarg = getopt_ordered(arg, long_opts)
  for i,v in ipairs (opts) do
    if v then
      if optarg [i] then
	if multi and ret[v] then
	  if type(ret[v]) ~= "table" then
	    ret[v] = {ret[v]}
	  end
	  ret[v][#ret[v]+1] = optarg[i]
	else
	  ret[v] = optarg[i]
	end
      else
        ret[v] = ret[v] and ret[v] + 1 or 1
      end
    else
      ret[#ret+1] = optarg[i]
    end
  end
  return ret
end
